@@ -34,37 +34,6 @@ import tempfile
 
 class SQLQuery():
     def __init__(self, api_key, instance_crn, target_cos_url=None, client_info=''):
-        self.endpoint_alias_mapping = {
-            "us-geo": "s3-api.us-geo.objectstorage.softlayer.net",
-            "us": "s3-api.us-geo.objectstorage.softlayer.net",
-            "dal-us-geo": "s3-api.dal-us-geo.objectstorage.softlayer.net",
-            "wdc-us-geo": "s3-api.wdc-us-geo.objectstorage.softlayer.net",
-            "sjc-us-geo": "s3-api.sjc-us-geo.objectstorage.softlayer.net",
-            "eu-geo": "s3.eu-geo.objectstorage.softlayer.net",
-            "eu": "s3.eu-geo.objectstorage.softlayer.net",
-            "ams-eu-geo": "s3.ams-eu-geo.objectstorage.softlayer.net",
-            "fra-eu-geo": "s3.fra-eu-geo.objectstorage.softlayer.net",
-            "mil-eu-geo": "s3.mil-eu-geo.objectstorage.softlayer.net",
-            "us-south": "s3.us-south.objectstorage.softlayer.net",
-            "us-east": "s3.us-east.objectstorage.softlayer.net",
-            "jp-tok": "s3.jp-tok.objectstorage.softlayer.net",
-            "ap-geo": "s3.ap-geo.objectstorage.softlayer.net",
-            "ap": "s3.ap-geo.objectstorage.softlayer.net",
-            "tok-ap-geo": "s3.tok-ap-geo.objectstorage.softlayer.net",
-            "seo-ap-geo": "s3.seo-ap-geo.objectstorage.softlayer.net",
-            "hkg-ap-geo": "s3.hkg-ap-geo.objectstorage.softlayer.net",
-            "eu-de": "s3.eu-de.objectstorage.softlayer.net",
-            "eu-gb": "s3.eu-gb.objectstorage.softlayer.net",
-            "ams03": "s3.ams03.objectstorage.softlayer.net",
-            "che01": "s3.che01.objectstorage.softlayer.net",
-            "mel01": "s3.mel01.objectstorage.softlayer.net",
-            "tor01": "s3.tor01.objectstorage.softlayer.net",
-            "mon01": "s3.mon01.objectstorage.softlayer.net",
-            "osl01": "s3.osl01.objectstorage.softlayer.net",
-            "sao01": "s3.sao01.objectstorage.softlayer.net",
-            "seo01": "s3.seo01.objectstorage.softlayer.net"
-        }
-        self.api_key = api_key
         self.instance_crn = instance_crn
         self.target_cos = target_cos_url
         self.export_cos_url = target_cos_url
@@ -82,34 +51,77 @@ class SQLQuery():
 
         self.logged_on = False
 
-    def logon(self):
-        if sys.version_info >= (3, 0):
-            data = urllib.parse.urlencode({'grant_type': 'urn:ibm:params:oauth:grant-type:apikey', 'apikey': self.api_key})
-        else:
-            data = urllib.urlencode({'grant_type': 'urn:ibm:params:oauth:grant-type:apikey', 'apikey': self.api_key})
+        # setup boto3 session in constructor to avoid leaking api key
+        ibm_boto3.setup_default_session(
+            ibm_api_key_id=api_key,
+        )
 
-        response = requests.post(
-            'https://iam.bluemix.net/identity/token',
-            headers=self.request_headers_xml_content,
-            data=data)
+    def _get_cos_client(self, endpoint=''):
+        '''
+        Create a service client of COS.
+        '''
+        return ibm_boto3.client('s3',
+            config=Config(signature_version='oauth'),
+            endpoint_url='https://{}'.format(endpoint)
+        )
 
-        if response.status_code == 200:
-            # print("Authentication successful")
-            bearer_response = response.json()
-            self.bearer_token = 'Bearer ' + bearer_response['access_token']
-            self.request_headers = {'Content-Type': 'application/json'}
-            self.request_headers.update({'Accept':'application/json'})
-            self.request_headers.update({'User-Agent': self.user_agent})
-            self.request_headers.update({'authorization': self.bearer_token})
-            self.logged_on = True
+    class ParsedUrl:
+        def __init__(self, url):
+            self.endpoint_alias_mapping = {
+                "us-geo": "s3-api.us-geo.objectstorage.softlayer.net",
+                "us": "s3-api.us-geo.objectstorage.softlayer.net",
+                "dal-us-geo": "s3-api.dal-us-geo.objectstorage.softlayer.net",
+                "wdc-us-geo": "s3-api.wdc-us-geo.objectstorage.softlayer.net",
+                "sjc-us-geo": "s3-api.sjc-us-geo.objectstorage.softlayer.net",
+                "eu-geo": "s3.eu-geo.objectstorage.softlayer.net",
+                "eu": "s3.eu-geo.objectstorage.softlayer.net",
+                "ams-eu-geo": "s3.ams-eu-geo.objectstorage.softlayer.net",
+                "fra-eu-geo": "s3.fra-eu-geo.objectstorage.softlayer.net",
+                "mil-eu-geo": "s3.mil-eu-geo.objectstorage.softlayer.net",
+                "us-south": "s3.us-south.objectstorage.softlayer.net",
+                "us-east": "s3.us-east.objectstorage.softlayer.net",
+                "jp-tok": "s3.jp-tok.objectstorage.softlayer.net",
+                "ap-geo": "s3.ap-geo.objectstorage.softlayer.net",
+                "ap": "s3.ap-geo.objectstorage.softlayer.net",
+                "tok-ap-geo": "s3.tok-ap-geo.objectstorage.softlayer.net",
+                "seo-ap-geo": "s3.seo-ap-geo.objectstorage.softlayer.net",
+                "hkg-ap-geo": "s3.hkg-ap-geo.objectstorage.softlayer.net",
+                "eu-de": "s3.eu-de.objectstorage.softlayer.net",
+                "eu-gb": "s3.eu-gb.objectstorage.softlayer.net",
+                "ams03": "s3.ams03.objectstorage.softlayer.net",
+                "che01": "s3.che01.objectstorage.softlayer.net",
+                "mel01": "s3.mel01.objectstorage.softlayer.net",
+                "tor01": "s3.tor01.objectstorage.softlayer.net",
+                "mon01": "s3.mon01.objectstorage.softlayer.net",
+                "osl01": "s3.osl01.objectstorage.softlayer.net",
+                "sao01": "s3.sao01.objectstorage.softlayer.net",
+                "seo01": "s3.seo01.objectstorage.softlayer.net"
+            }
+            self.endpoint = url.split("/")[2]
+            self.endpoint = self.endpoint_alias_mapping.get(self.endpoint, self.endpoint)
+            self.bucket = url.split("/")[3]
+            self.prefix = url[url.replace('/', 'X', 3).find('/') + 1:]
+            if len(self.prefix) > 0 and self.prefix[-1] == '*':
+                self.prefix = self.prefix[:-1]
+                self.fourth_slash = url.replace('/', 'X', 3).find('/')
 
-        else:
-            print("Authentication failed with http code {}".format(response.status_code))
+    def logon(self, force=False):
+        if self.logged_on and not force and (datetime.now() - self.last_logon).seconds < 300:
+            return
+
+        ## TODO refactor construction to avoid calling private method
+        boto3_session = ibm_boto3._get_default_session()
+        ro_credentials = boto3_session.get_credentials().get_frozen_credentials()
+
+        self.request_headers = {'Content-Type': 'application/json'}
+        self.request_headers.update({'Accept':'application/json'})
+        self.request_headers.update({'User-Agent': self.user_agent})
+        self.request_headers.update({'authorization': 'Bearer {}'.format(ro_credentials.token)})
+        self.logged_on = True
+        self.last_logon = datetime.now()
 
     def submit_sql(self, sql_text, pagesize=None):
-        if not self.logged_on:
-            print("You are not logged on to IBM Cloud")
-            return
+        self.logon()
         sqlData = {'statement': sql_text}
         # If a valid pagesize is specified we need to append the proper PARTITIONED EVERY <num> ROWS clause
         if pagesize or pagesize==0:
@@ -128,22 +140,23 @@ class SQLQuery():
 
         try:
             response = requests.post(
-                "https://sql-api.ng.bluemix.net/v2/sql_jobs?instance_crn={}".format(self.instance_crn),
+                "https://api.sql-query.cloud.ibm.com/v2/sql_jobs?instance_crn={}".format(self.instance_crn),
                 headers=self.request_headers,
                 json=sqlData)
             resp = response.json()
             return resp['job_id']
+        except KeyError as e:
+            raise SyntaxError("SQL submission failed: {}".format(response.json()['errors'][0]['message']))
+
         except HTTPError as e:
             raise SyntaxError("SQL submission failed: {}".format(response.json()['errors'][0]['message']))
 
     def wait_for_job(self, jobId):
-        if not self.logged_on:
-            print("You are not logged on to IBM Cloud")
-            return "Not logged on"
+        self.logon()
 
         while True:
             response = requests.get(
-                "https://sql-api.ng.bluemix.net/v2/sql_jobs/{}?instance_crn={}".format(jobId, self.instance_crn),
+                "https://api.sql-query.cloud.ibm.com/v2/sql_jobs/{}?instance_crn={}".format(jobId, self.instance_crn),
                 headers=self.request_headers,
             )
 
@@ -151,8 +164,6 @@ class SQLQuery():
                 status_response = response.json()
                 jobStatus = status_response['status']
                 if jobStatus == 'completed':
-                    # print("Job {} has completed successfully".format(jobId))
-                    resultset_location = status_response['resultset_location']
                     break
                 if jobStatus == 'failed':
                     print("Job {} has failed".format(jobId))
@@ -167,26 +178,23 @@ class SQLQuery():
         return 0
 
     def get_result(self, jobId, pagenumber=None):
-        if not self.logged_on:
-            print("You are not logged on to IBM Cloud")
-            return
+        self.logon()
 
         job_details = self.get_job(jobId)
         job_status = job_details.get('status')
         if job_status == 'running':
-            raise ValueError('SQL job with jobId {} still running. Come back later.')
+            raise ValueError('SQL job with jobId {} still running. Come back later.'.format(jobId))
         elif job_status != 'completed':
-            raise ValueError('SQL job with jobId {} did not finish successfully. No result available.')
+            raise ValueError('SQL job with jobId {} did not finish successfully. No result available.'.format(jobId))
 
-        result_cos_url = job_details['resultset_location']
-        provided_cos_endpoint = result_cos_url.split("/")[2]
-        result_cos_endpoint = self.endpoint_alias_mapping.get(provided_cos_endpoint, provided_cos_endpoint)
-        result_cos_bucket = result_cos_url.split("/")[3]
-        result_cos_prefix = result_cos_url[result_cos_url.replace('/', 'X', 3).find('/')+1:]
-        result_location = "https://{}/{}?prefix={}".format(result_cos_endpoint, result_cos_bucket, result_cos_prefix)
+        if "resultset_location" not in job_details:
+            return None
+
+        url_parsed = self.ParsedUrl(job_details['resultset_location'])
+        result_location = "https://{}/{}?prefix={}".format(url_parsed.endpoint, url_parsed.bucket, url_parsed.prefix)
         result_format = job_details['resultset_format']
 
-        if result_format not in ["csv", "parquet"]:
+        if result_format not in ["csv", "parquet", "json"]:
             raise ValueError("Result object format {} currently not supported by get_result().".format(result_format))
 
         response = requests.get(
@@ -208,11 +216,7 @@ class SQLQuery():
             raise ValueError("Result object listing for job {} at {} failed with http code {}".format(jobId, result_location,
                                                                                            response.status_code))
 
-        cos_client = ibm_boto3.client(service_name='s3',
-                                      ibm_api_key_id=self.api_key,
-                                      ibm_auth_endpoint="https://iam.ng.bluemix.net/oidc/token",
-                                      config=Config(signature_version='oauth'),
-                                      endpoint_url='https://' + result_cos_endpoint)
+        cos_client = self._get_cos_client(url_parsed.endpoint)
 
         # When pagenumber is specified we only retrieve that page. Otherwise we concatenate all pages to one DF:
         if pagenumber or pagenumber==0:
@@ -220,15 +224,20 @@ class SQLQuery():
                 raise ValueError("pagenumber ({}) specified, but the job was not submitted with pagination option.".format(pagenumber))
             if type(pagenumber) == int and 0 < pagenumber <= len(bucket_objects):
                 if result_format == "csv":
-                    body = cos_client.get_object(Bucket=result_cos_bucket, Key=bucket_objects[pagenumber-1])['Body']
+                    body = cos_client.get_object(Bucket=url_parsed.bucket, Key=bucket_objects[pagenumber-1])['Body']
                     if not hasattr(body, "__iter__"): body.__iter__ = types.MethodType(self.__iter__, body)
                     result_df = pd.read_csv(body)
                 elif result_format == "parquet":
                     tmpfile = tempfile.NamedTemporaryFile()
                     tempfilename = tmpfile.name
                     tmpfile.close()
-                    cos_client.download_file(Bucket=result_cos_bucket, Key=bucket_objects[pagenumber-1], Filename=tempfilename)
+                    cos_client.download_file(Bucket=url_parsed.bucket, Key=bucket_objects[pagenumber-1], Filename=tempfilename)
                     result_df = pd.read_parquet(tempfilename)
+                elif result_format == "json":
+                    body = cos_client.get_object(Bucket=url_parsed.bucket, Key=bucket_objects[pagenumber-1])['Body']
+                    body = body.read().decode('utf-8')
+                    result_df = pd.read_json(body,lines=True)
+                    
             else:
                 raise ValueError("Invalid pagenumner ({}) specified".format(pagenumber))
         else:
@@ -236,7 +245,7 @@ class SQLQuery():
             for bucket_object in bucket_objects:
 
                 if result_format == "csv":
-                    body = cos_client.get_object(Bucket=result_cos_bucket, Key=bucket_object)['Body']
+                    body = cos_client.get_object(Bucket=url_parsed.bucket, Key=bucket_object)['Body']
                     # add missing __iter__ method, so pandas accepts body as file-like object
                     if not hasattr(body, "__iter__"): body.__iter__ = types.MethodType(self.__iter__, body)
 
@@ -246,12 +255,18 @@ class SQLQuery():
                     tmpfile = tempfile.NamedTemporaryFile()
                     tempfilename = tmpfile.name
                     tmpfile.close()
-                    cos_client.download_file(Bucket=result_cos_bucket, Key=bucket_object, Filename=tempfilename)
+                    cos_client.download_file(Bucket=url_parsed.bucket, Key=bucket_object, Filename=tempfilename)
 
                     partition_df = pd.read_parquet(tempfilename)
 
+                elif result_format == "json":
+                    body = cos_client.get_object(Bucket=url_parsed.bucket, Key=bucket_object)['Body']
+                    body = body.read().decode('utf-8')
+
+                    partition_df = pd.read_json(body, lines=True)
+                    
                 # Add columns from hive style partition naming schema
-                hive_partition_candidates = bucket_object.replace(result_cos_prefix + '/', '').split('/')
+                hive_partition_candidates = bucket_object.replace(url_parsed.prefix + '/', '').split('/')
                 for hive_partition_candidate in hive_partition_candidates:
                     if hive_partition_candidate.count('=') == 1: # Hive style folder names contain exactly one '='
                         column = hive_partition_candidate.split('=')
@@ -265,7 +280,7 @@ class SQLQuery():
                 if 'result_df' not in locals():
                     result_df = partition_df
                 else:
-                    result_df = result_df.append(partition_df)
+                    result_df = result_df.append(partition_df, sort=False)
 
             if 'result_df' not in locals():
                 return None
@@ -273,9 +288,7 @@ class SQLQuery():
         return result_df
 
     def list_results(self, jobId):
-        if not self.logged_on:
-            print("You are not logged on to IBM Cloud")
-            return
+        self.logon()
 
         job_details = self.get_job(jobId)
         if job_details['status'] == 'running':
@@ -284,8 +297,7 @@ class SQLQuery():
             raise ValueError('SQL job with jobId {} did not finish successfully. No result available.')
 
         result_location = job_details['resultset_location'].replace("cos", "https", 1)
-        provided_cos_endpoint = job_details['resultset_location'].split("/")[2]
-        result_cos_endpoint = self.endpoint_alias_mapping.get(provided_cos_endpoint, provided_cos_endpoint)
+        url_parsed = self.ParsedUrl(result_location)
 
         fourth_slash = result_location.replace('/', 'X', 3).find('/')
 
@@ -296,15 +308,15 @@ class SQLQuery():
 
         if response.status_code == 200 or response.status_code == 201:
             ns = {'s3': 'http://s3.amazonaws.com/doc/2006-03-01/'}
-            responseBodyXMLroot = ET.fromstring(response.body)
+            responseBodyXMLroot = ET.fromstring(response.content)
             bucket_name = responseBodyXMLroot.find('s3:Name', ns).text
             bucket_objects = []
             if responseBodyXMLroot.findall('s3:Contents', ns):
                 for contents in responseBodyXMLroot.findall('s3:Contents', ns):
                     key = contents.find('s3:Key', ns)
-                    object_url = "cos://{}/{}/{}".format(result_cos_endpoint, bucket_name, key.text)
+                    object_url = "cos://{}/{}/{}".format(url_parsed.endpoint, bucket_name, key.text)
                     size = contents.find('s3:Size', ns)
-                    bucket_objects.append({'Key': object_url, 'Size': size.text})
+                    bucket_objects.append({'Key': object_url, 'Size': size.text, 'Bucket': bucket_name, 'Object': key.text})
             else:
                 print('There are no result objects for the jobid {}'.format(jobId))
                 return
@@ -315,14 +327,42 @@ class SQLQuery():
 
         result_objects_df = pd.DataFrame(columns=['ObjectURL', 'Size'])
         for object in bucket_objects:
-            result_objects_df = result_objects_df.append([{'ObjectURL': object['Key'], 'Size': object['Size']}], ignore_index=True)
+            result_objects_df = result_objects_df.append([{'ObjectURL': object['Key'],
+                                                           'Size': object['Size'],
+                                                           'Bucket': object['Bucket'],
+                                                           'Object': object['Object']}], ignore_index=True, sort=False)
 
         return result_objects_df
 
+    def rename_exact_result(self, jobId):
+        self.logon()
+
+        job_details = self.get_job(jobId)
+        job_status = job_details.get('status')
+        if job_status == 'running':
+            raise ValueError('SQL job with jobId {} still running. Come back later.'.format(jobId))
+        elif job_status != 'completed':
+            raise ValueError('SQL job with jobId {} did not finish successfully. No result available.'.format(jobId))
+
+        url_parsed = self.ParsedUrl(job_details['resultset_location'])
+        cos_client = self._get_cos_client(url_parsed.endpoint)
+
+        result_objects = self.list_results(jobId)
+
+        if len(result_objects) > 3:
+            raise ValueError('Renaming partitioned results of jobId {} to single exact result object name not supported.'.format(jobId))
+        if len(result_objects) < 3 or int(result_objects.Size[0]) != 0 or int(result_objects.Size[1]) != 0:
+            raise ValueError('Results of jobId {} don\'t seem to be regular SQL query output.'.format(jobId))
+
+        copy_source = result_objects.Bucket[2] + "/" + result_objects.Object[2]
+        cos_client.copy_object(Bucket=result_objects.Bucket[0], CopySource=copy_source, Key=result_objects.Object[0])
+        cos_client.delete_object(Bucket=result_objects.Bucket[2], Key=result_objects.Object[2])
+        cos_client.delete_object(Bucket=result_objects.Bucket[1], Key=result_objects.Object[1])
+
+        return
+
     def delete_result(self, jobId):
-        if not self.logged_on:
-            print("You are not logged on to IBM Cloud")
-            return
+        self.logon()
 
         job_details = self.get_job(jobId)
         if job_details['status'] == 'running':
@@ -331,8 +371,7 @@ class SQLQuery():
             raise ValueError('SQL job with jobId {} did not finish successfully. No result available.')
 
         result_location = job_details['resultset_location'].replace("cos", "https", 1)
-        provided_cos_endpoint = job_details['resultset_location'].split("/")[2]
-        result_cos_endpoint = self.endpoint_alias_mapping.get(provided_cos_endpoint, provided_cos_endpoint)
+        url_parsed = self.ParsedUrl(result_location)
 
         fourth_slash = result_location.replace('/', 'X', 3).find('/')
 
@@ -343,7 +382,7 @@ class SQLQuery():
 
         if response.status_code == 200 or response.status_code == 201:
             ns = {'s3': 'http://s3.amazonaws.com/doc/2006-03-01/'}
-            responseBodyXMLroot = ET.fromstring(response.body)
+            responseBodyXMLroot = ET.fromstring(response.content)
             bucket_name = responseBodyXMLroot.find('s3:Name', ns).text
             bucket_objects = []
             if responseBodyXMLroot.findall('s3:Contents', ns):
@@ -358,28 +397,22 @@ class SQLQuery():
                                                                                            response.status_code))
             return
 
-        cos_client = ibm_boto3.client(service_name='s3',
-                                      ibm_api_key_id=self.api_key,
-                                      ibm_auth_endpoint="https://iam.ng.bluemix.net/oidc/token",
-                                      config=Config(signature_version='oauth'),
-                                      endpoint_url='https://' + result_cos_endpoint)
+        cos_client = self._get_cos_client(url_parsed.endpoint)
 
         response = cos_client.delete_objects(Bucket=bucket_name, Delete={'Objects': bucket_objects})
 
         deleted_list_df = pd.DataFrame(columns=['Deleted Object'])
         for deleted_object in response['Deleted']:
-            deleted_list_df = deleted_list_df.append([{'Deleted Object': deleted_object['Key']}], ignore_index=True)
+            deleted_list_df = deleted_list_df.append([{'Deleted Object': deleted_object['Key']}], ignore_index=True, sort=False)
 
         return deleted_list_df
 
     def get_job(self, jobId):
-        if not self.logged_on:
-            print("You are not logged on to IBM Cloud")
-            return
+        self.logon()
 
         try:
             response = requests.get(
-                "https://sql-api.ng.bluemix.net/v2/sql_jobs/{}?instance_crn={}".format(jobId, self.instance_crn),
+                "https://api.sql-query.cloud.ibm.com/v2/sql_jobs/{}?instance_crn={}".format(jobId, self.instance_crn),
                 headers=self.request_headers,
             )
         except HTTPError as e:
@@ -391,12 +424,10 @@ class SQLQuery():
         return response.json()
 
     def get_jobs(self):
-        if not self.logged_on:
-            print("You are not logged on to IBM Cloud")
-            return
+        self.logon()
 
         response = requests.get(
-            "https://sql-api.ng.bluemix.net/v2/sql_jobs?instance_crn={}".format(self.instance_crn),
+            "https://api.sql-query.cloud.ibm.com/v2/sql_jobs?instance_crn={}".format(self.instance_crn),
             headers=self.request_headers,
             )
         if response.status_code == 200 or response.status_code == 201:
@@ -406,7 +437,7 @@ class SQLQuery():
                                                 'error', 'error_message'])
             for job in job_list['jobs']:
                 response = requests.get(
-                    "https://sql-api.ng.bluemix.net/v2/sql_jobs/{}?instance_crn={}".format(job['job_id'],
+                    "https://api.sql-query.cloud.ibm.com/v2/sql_jobs/{}?instance_crn={}".format(job['job_id'],
                                                                                                 self.instance_crn),
                     headers=self.request_headers,
                     )
@@ -442,7 +473,7 @@ class SQLQuery():
                                                        'bytes_read': bytes_read,
                                                        'error': error,
                                                        'error_message': error_message,
-                                                       }], ignore_index=True)
+                                                       }], ignore_index=True, sort=False)
                 else:
                     print("Job details retrieval for jobId {} failed with http code {}".format(job['job_id'],
                                                                                                response.status_code))
@@ -465,15 +496,13 @@ class SQLQuery():
             return self.get_result(jobId)
 
     def sql_ui_link(self):
-        if not self.logged_on:
-            print("You are not logged on to IBM Cloud")
-            return
+        self.logon()
 
         if sys.version_info >= (3, 0):
-            print ("https://sql.ng.bluemix.net/sqlquery/?instance_crn={}".format(
+            print ("https://sql-query.cloud.ibm.com/sqlquery/?instance_crn={}".format(
                 urllib.parse.unquote(self.instance_crn)))
         else:
-            print ("https://sql.ng.bluemix.net/sqlquery/?instance_crn={}".format(
+            print ("https://sql-query.cloud.ibm.com/sqlquery/?instance_crn={}".format(
                 urllib.unquote(self.instance_crn).decode('utf8')))
 
     def get_cos_summary(self, url):
@@ -484,23 +513,13 @@ class SQLQuery():
                 num /= 1024.0
             return "%.1f %s%s" % (num, 'Y', suffix)
 
-        if not self.logged_on:
-            print("You are not logged on to IBM Cloud")
-            return
+        self.logon()
 
-        endpoint = url.split("/")[2]
-        endpoint = self.endpoint_alias_mapping.get(endpoint, endpoint)
-        bucket = url.split("/")[3]
-        prefix = url[url.replace('/', 'X', 3).find('/') + 1:]
-
-        cos_client = ibm_boto3.client(service_name='s3',
-                                      ibm_api_key_id=self.api_key,
-                                      ibm_auth_endpoint="https://iam.ng.bluemix.net/oidc/token",
-                                      config=Config(signature_version='oauth'),
-                                      endpoint_url='https://' + endpoint)
+        url_parsed = self.ParsedUrl(url)
+        cos_client = self._get_cos_client(url_parsed.endpoint)
 
         paginator = cos_client.get_paginator("list_objects")
-        page_iterator = paginator.paginate(Bucket=bucket, Prefix=prefix)
+        page_iterator = paginator.paginate(Bucket=url_parsed.bucket, Prefix=url_parsed.prefix)
 
         total_size = 0
         smallest_size = 9999999999999999
@@ -543,6 +562,25 @@ class SQLQuery():
                 'smallest_object_size': sizeof_fmt(smallest_size), 'smallest_object': smallest_object,
                 'largest_object_size': sizeof_fmt(largest_size), 'largest_object': largest_object}
 
+    def list_cos_objects(self, url):
+        self.logon()
+
+        url_parsed = self.ParsedUrl(url)
+        cos_client = self._get_cos_client(url_parsed.endpoint)
+        paginator = cos_client.get_paginator("list_objects")
+        page_iterator = paginator.paginate(Bucket=url_parsed.bucket, Prefix=url_parsed.prefix)
+
+        for page in page_iterator:
+            if "Contents" in page:
+                #print(page['Contents'])
+                page_df = pd.DataFrame.from_dict(page['Contents'], orient='columns')
+                if 'result' in locals():
+                    result = result.append(page_df, sort=False)
+                else:
+                    result = page_df
+        result = result.drop(columns=['ETag', 'Owner']).rename(columns={"Key": "Object"})
+        return result
+
     def export_job_history(self, cos_url=None):
         if cos_url:
             # Default export location is target COS URL set at __init__
@@ -552,10 +590,7 @@ class SQLQuery():
             raise ValueError('No configured export COS URL.')
         if not self.export_cos_url.endswith('/'):
             self.export_cos_url += "/"
-        provided_export_cos_endpoint = self.export_cos_url.split("/")[2]
-        export_cos_endpoint = self.endpoint_alias_mapping.get(provided_export_cos_endpoint, provided_export_cos_endpoint)
-        export_cos_bucket = self.export_cos_url.split("/")[3]
-        export_cos_prefix = self.export_cos_url[self.export_cos_url.replace('/', 'X', 3).find('/')+1:]
+        url_parsed = self.ParsedUrl(self.export_cos_url)
         export_file_prefix = "job_export_"
 
         job_history_df = self.get_jobs() # Retrieve current job history (most recent 30 jobs)
@@ -563,20 +598,16 @@ class SQLQuery():
         newest_job_end_time = terminated_job_history_df.loc[pd.to_datetime(terminated_job_history_df['end_time']).idxmax()].end_time
 
         # List all existing objects in export location and identify latest exported job timestamp:
-        cos_client = ibm_boto3.client(service_name='s3',
-                                      ibm_api_key_id=self.api_key,
-                                      ibm_auth_endpoint="https://iam.ng.bluemix.net/oidc/token",
-                                      config=Config(signature_version='oauth'),
-                                      endpoint_url='https://' + export_cos_endpoint)
+        cos_client = self._get_cos_client(url_parsed.endpoint)
         paginator = cos_client.get_paginator("list_objects")
-        page_iterator = paginator.paginate(Bucket=export_cos_bucket, Prefix=export_cos_prefix)
+        page_iterator = paginator.paginate(Bucket=url_parsed.bucket, Prefix=url_parsed.prefix)
         newest_exported_job_end_time = ""
         for page in page_iterator:
             if "Contents" in page:
                 for key in page['Contents']:
                     object_name = key["Key"]
                     suffix_index = object_name.find(".parquet")
-                    prefix_end_index = len(export_cos_prefix + export_file_prefix)
+                    prefix_end_index = len(url_parsed.prefix + export_file_prefix)
                     if prefix_end_index < suffix_index:
                         job_end_time = object_name[prefix_end_index:suffix_index]
                         if job_end_time > newest_exported_job_end_time:
@@ -588,7 +619,7 @@ class SQLQuery():
             tempfilename = tmpfile.name
             new_jobs_df = terminated_job_history_df[terminated_job_history_df['end_time'] > newest_exported_job_end_time]
             new_jobs_df.to_parquet(engine="pyarrow", fname=tempfilename, compression="snappy")
-            cos_client.upload_file(Bucket=export_cos_bucket, Filename=tempfilename, Key=export_cos_prefix + export_file_prefix + newest_job_end_time + ".parquet")
+            cos_client.upload_file(Bucket=url_parsed.bucket, Filename=tempfilename, Key=url_parsed.prefix + export_file_prefix + newest_job_end_time + ".parquet")
             print("Exported {} new jobs".format(new_jobs_df['job_id'].count()))
             tmpfile.close()
         else:
